from dataclasses import dataclass
from typing import List


@dataclass
class Dialect:
    KEYWORDS: List[str]

    NULL: bytes
    ONE: bytes
    TWO: bytes
    Q_KW: bytes
    A_KW: bytes
    C_KW: bytes


CHIA_DIALECT = Dialect(
    (
        # core opcodes 0x01-x08
        ". q a i c f r l x "
        # opcodes on atoms as strings 0x09-0x0f
        "= >s sha256 substr strlen concat . "
        # opcodes on atoms as ints 0x10-0x17
        "+ - * / divmod > ash lsh "
        # opcodes on atoms as vectors of bools 0x18-0x1c
        "logand logior logxor lognot . "
        # opcodes for bls 1381 0x1d-0x1f
        "point_add pubkey_for_exp . "
        # bool opcodes 0x20-0x23
        "not any all . "
        # misc 0x24
        "softfork "
    ).split(),
    NULL=bytes.fromhex(""),
    ONE=bytes.fromhex("01"),
    TWO=bytes.fromhex("02"),
    Q_KW=bytes.fromhex("01"),
    A_KW=bytes.fromhex("02"),
    C_KW=bytes.fromhex("04"),
)



"""
Some utilities to cast python types to and from clvm.
"""

from typing import Callable, List, SupportsBytes, Tuple, Union, cast

from .clvm_storage import CLVMStorage, is_clvm_storage


AtomCastableType = Union[
    bytes,
    str,
    int,
    SupportsBytes,
]


CastableType = Union[
    AtomCastableType,
    List["CastableType"],
    Tuple["CastableType", "CastableType"],
    CLVMStorage,
]


NULL_BLOB = b""


def generate_working_memview_or_bytes():
    # in python3.7 you can't go 'isinstance(b"", SupportsBytes)'

    # this one works in py38+
    def memview_or_bytes_py38_or_later(o):
        return isinstance(o, (memoryview, SupportsBytes))

    # this one works in py37
    def memview_or_bytes_py37(o):
        return getattr(o, "__bytes__", None) is not None

    try:
        memview_or_bytes_py38_or_later(b"")
        return memview_or_bytes_py38_or_later
    except TypeError:
        pass
    return memview_or_bytes_py37


memview_or_bytes = generate_working_memview_or_bytes()


def int_from_bytes(blob: bytes) -> int:
    """
    Convert a bytes blob encoded as a clvm int to a python int.
    """
    size = len(blob)
    if size == 0:
        return 0
    return int.from_bytes(blob, "big", signed=True)


def int_to_bytes(v: int) -> bytes:
    """
    Convert a python int to a blob that encodes as this integer in clvm.
    """
    if v == 0:
        return b""
    byte_count = (v.bit_length() + 8) // 8
    r = v.to_bytes(byte_count, "big", signed=True)
    # make sure the string returned is minimal
    # ie. no leading 00 or ff bytes that are unnecessary
    while len(r) > 1 and r[0] == (0xFF if r[1] & 0x80 else 0):
        r = r[1:]
    return r


def to_atom_type(v: AtomCastableType) -> bytes:
    """
    Convert an `AtomCastableType` to `bytes`. This for use with the
    convenience function `Program.to`.
    """
    if isinstance(v, bytes):
        return v
    if isinstance(v, str):
        return v.encode()
    if isinstance(v, int):
        return int_to_bytes(v)
    if memview_or_bytes(v):
        return bytes(v)

    raise ValueError("can't cast %s (%s) to bytes" % (type(v), v))


def to_clvm_object(
    castable: CastableType,
    to_atom_f: Callable[[bytes], CLVMStorage],
    to_pair_f: Callable[[CLVMStorage, CLVMStorage], CLVMStorage],
) -> CLVMStorage:
    """
    Convert a python object to clvm object.

    This works on nested tuples and lists of potentially unlimited depth.
    It is non-recursive, so nesting depth is not limited by the call stack.
    But the entire hiearachy must be traversed, so execution time is
    proportional to hierarchy depth, and thus potentially unbounded.

    So don't use on untrusted input. Also, the case where a list that transitively
    contains itself (eg `t = []; t.append(t)`) in a loop is not detected, and
    this function will never return because it acts like a hiearachy with
    infinite depth.
    """
    to_convert: List[CastableType] = [castable]
    did_convert: List[CLVMStorage] = []
    ops: List[int] = [0]

    # operations:
    #  0: pop `to_convert` and convert if possible, storing result on `did_convert`,
    #     or subdivide task, pushing multiple things on `to_convert` (and new ops)
    #  1: pop & cons two items from `did_convert`, pushing result to `did_convert`
    #  2: same as 1 but cons in opposite order. Necessary for converting lists

    while len(ops) > 0:
        op = ops.pop()
        # convert value
        if op == 0:
            v = to_convert.pop()
            if is_clvm_storage(v):
                v = cast(CLVMStorage, v)
                did_convert.append(v)
                continue
            if isinstance(v, tuple):
                if len(v) != 2:
                    raise ValueError("can't cast tuple of size %d" % len(v))
                left, right = v
                ll_right = is_clvm_storage(right)
                ll_left = is_clvm_storage(left)
                if ll_right and ll_left:
                    left = cast(CLVMStorage, left)
                    right = cast(CLVMStorage, right)
                    did_convert.append(to_pair_f(left, right))
                else:
                    ops.append(1)  # cons
                    to_convert.append(left)
                    ops.append(0)  # convert
                    to_convert.append(right)
                    ops.append(0)  # convert
                continue
            if isinstance(v, list):
                for _ in v:
                    ops.append(2)  # rcons

                # add and convert the null terminator
                to_convert.append(to_atom_f(NULL_BLOB))
                ops.append(0)  # convert

                for _ in reversed(v):
                    to_convert.append(_)
                    ops.append(0)  # convert
                continue
            v = cast(AtomCastableType, v)
            did_convert.append(to_atom_f(to_atom_type(v)))
            continue
        if op == 1:  # cons
            left = did_convert.pop()
            right = did_convert.pop()
            obj = to_pair_f(left, right)
            did_convert.append(obj)
            continue
        if op == 2:  # rcons
            right = did_convert.pop()
            left = did_convert.pop()
            obj = to_pair_f(left, right)
            did_convert.append(obj)
            continue

    # there's exactly one item left at this point
    assert len(did_convert) == 1
    return did_convert[0]



from __future__ import annotations
from typing import Iterator, List, Tuple, Optional, BinaryIO

from .at import at
from .casts import CastableType, to_clvm_object, int_from_bytes, int_to_bytes
from .chia_dialect import CHIA_DIALECT
from .clvm_rs import run_serialized_chia_program
from .clvm_storage import CLVMStorage
from .clvm_tree import CLVMTree
from .curry_and_treehash import CurryTreehasher
from .eval_error import EvalError
from .replace import replace
from .ser import sexp_from_stream, sexp_to_stream, sexp_to_bytes
from .tree_hash import sha256_treehash



class Program(CLVMStorage):
    """
    A wrapper around `CLVMStorage` providing many convenience functions.
    """

    UNSAFE_MAX_COST: Optional[int] = None

    curry_treehasher: CurryTreehasher = CurryTreehasher(CHIA_DIALECT)
    _cached_serialization: Optional[bytes]

    # serialization/deserialization

    @classmethod
    def parse(cls, f: BinaryIO) -> Program:
        return sexp_from_stream(f, cls.new_pair, cls.new_atom)

    def stream(self, f: BinaryIO) -> None:
        sexp_to_stream(self, f)

    @classmethod
    def from_bytes(cls, blob: bytes, calculate_tree_hash: bool = True) -> Program:
        obj, cursor = cls.from_bytes_with_cursor(
            blob, 0, calculate_tree_hash=calculate_tree_hash
        )
        return obj

    @classmethod
    def from_bytes_with_cursor(
        cls, blob: bytes, cursor: int, calculate_tree_hash: bool = True
    ) -> Tuple[Program, int]:
        tree = CLVMTree.from_bytes(
            blob[cursor:], calculate_tree_hash=calculate_tree_hash
        )
        obj = cls.wrap(tree)
        new_cursor = len(bytes(tree)) + cursor
        return obj, new_cursor

    @classmethod
    def fromhex(cls, hexstr: str) -> Program:
        return cls.from_bytes(bytes.fromhex(hexstr))

    def __bytes__(self) -> bytes:
        if self._cached_serialization is None:
            self._cached_serialization = sexp_to_bytes(self)
        if not isinstance(self._cached_serialization, bytes):
            self._cached_serialization = bytes(self._cached_serialization)
        return self._cached_serialization

    def __int__(self) -> int:
        v = self.as_int()
        if v is None:
            raise ValueError("can't cast pair to int")
        return v

    def __hash__(self):
        return id(self)

    @classmethod
    def int_from_bytes(cls, b: bytes) -> int:
        return int_from_bytes(b)

    @classmethod
    def int_to_bytes(cls, i: int) -> bytes:
        return int_to_bytes(i)

    # high level casting with `.to`

    def __init__(self):
        self.atom = b""
        self._pair = None
        self._unwrapped = self
        self._unwrapped_pair = None
        self._cached_serialization = None
        self._cached_sha256_treehash = None

    @property
    def pair(self) -> Optional[Tuple["Program", "Program"]]:
        if self._pair is None and self.atom is None:
            pair = self._unwrapped_pair
            self._pair = (self.wrap(pair[0]), self.wrap(pair[1]))
        return self._pair

    @classmethod
    def to(cls, v: CastableType) -> Program:
        return cls.wrap(to_clvm_object(v, cls.new_atom, cls.new_pair))

    @classmethod
    def wrap(cls, v: CLVMStorage) -> Program:
        if isinstance(v, Program):
            return v
        o = cls()
        o.atom = v.atom
        o._pair = None
        o._unwrapped = v
        o._unwrapped_pair = v.pair
        o._cached_serialization = getattr(v, "_cached_serialization", None)
        o._cached_sha256_treehash = getattr(v, "_cached_sha256_treehash", None)
        return o

    # new object creation on the python heap

    @classmethod
    def new_atom(cls, v: bytes) -> Program:
        o = cls()
        o.atom = bytes(v)
        o._pair = None
        o._unwrapped_pair = None
        return o

    @classmethod
    def new_pair(cls, left: CLVMStorage, right: CLVMStorage) -> Program:
        o = cls()
        o.atom = None
        o._pair = None
        o._unwrapped_pair = (left, right)
        return o

    @classmethod
    def null(cls) -> Program:
        return NULL_PROGRAM

    @classmethod
    def one(cls) -> Program:
        return ONE_PROGRAM

    # display

    def __str__(self) -> str:
        s = bytes(self).hex()
        if len(s) > 76:
            s = f"{s[:70]}...{s[-6:]}"
        return s

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self)})"

    def __eq__(self, other) -> bool:
        try:
            other_obj = self.to(other)
        except ValueError:
            # cast failure
            return False
        return self.tree_hash() == other_obj.tree_hash()

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def first(self) -> Optional[Program]:
        if self.pair:
            return self.pair[0]
        return None

    def rest(self) -> Optional[Program]:
        if self.pair:
            return self.pair[1]
        return None

    def listp(self) -> bool:
        return self.pair is not None

    def nullp(self) -> bool:
        return self.atom == b""

    def list_len(self) -> int:
        c = 0
        v: CLVMStorage = self
        while v.pair is not None:
            v = v.pair[1]
            c += 1
        return c

    def at(self, position: str) -> Optional["Program"]:
        """
        Take a string of `f` and `r` characters and follow that path.

        Example:

        ```
        p1 = Program.to([10, 20, 30, [15, 17], 40, 50])
        assert Program.to(17) == at(p1, "rrrfrf")
        ```

        Returns `None` if an atom is hit at some intermediate node.

        ```
        p1 = Program.to(10)
        assert None == at(p1, "rr")
        ```
        """
        r = at(self, position)
        if r is None:
            return r
        return self.to(r)

    def at_many(self, *positions: str) -> List[Optional["Program"]]:
        """
        Call `.at` multiple times.

        Why? So you can write

        `
        if p.at_many("f", "rf", "rfrf") == [5, 10, 15]:
        `
        instead of
        `
        if [p.at("f"), p.at("rf"), p.at("rfrf")] == [5, 10, 15]:
        `
        """
        return [self.at(_) for _ in positions]

    def replace(self, **kwargs) -> "Program":
        """
        Create a new program replacing the given paths (using `at` syntax).
        Example:
        ```
        >>> p1 = Program.to([100, 200, 300])
        >>> print(p1.replace(f=105) == Program.to([105, 200, 300]))
        True
        >>> p2 = [100, 200, [301, 302]]
        >>> print(p1.replace(rrf=[301, 302]) == Program.to(p2))
        True
        >>> p2 = [105, 200, [301, 302]]
        >>> print(p1.replace(f=105, rrf=[301, 302]) == Program.to(p2))
        True
        ```

        This is a convenience method intended for use in the wallet or
        command-line hacks where it would be easier to morph elements
        of an existing clvm object tree than to rebuild one from scratch.

        Note that `Program` objects are immutable. This function returns a
        new object; the original is left as-is.
        """
        return self.to(replace(self, **kwargs))

    def tree_hash(self) -> bytes:
        # we operate on the unwrapped version to prevent the re-wrapping that
        # happens on each invocation of `Program.pair` whenever possible
        if self._cached_sha256_treehash is None:
            self._cached_sha256_treehash = sha256_treehash(self._unwrapped)
        return self._cached_sha256_treehash

    def run_with_cost(
        self, args, max_cost: int, flags: int = 0
    ) -> Tuple[int, "Program"]:
        prog_bytes = bytes(self)
        args_bytes = bytes(self.to(args))
        try:
            cost, lazy_node = run_serialized_chia_program(
                prog_bytes, args_bytes, max_cost, flags
            )
            r = self.wrap(lazy_node)
        except ValueError as ve:
            raise EvalError(ve.args[0], self.wrap(ve.args[1]))
        return cost, r

    def run(self, args) -> "Program":
        """
        Run with the default `UNSAFE_MAX_COST` value. Using too high a value with
        misbehaving code may exhaust memory or take a long time.
        """
        max_cost = self.__class__.UNSAFE_MAX_COST
        if max_cost is None:
            raise ValueError("please call `set_run_unsafe_max_cost` before using `run`")
        cost, r = self.run_with_cost(args, max_cost=max_cost)
        return r

    @classmethod
    def set_run_unsafe_max_cost(cls, new_max_cost: int):
        cls.UNSAFE_MAX_COST = new_max_cost

    def curry(self, *args: CastableType) -> "Program":
        """
        Given a `MOD` program, cast to `Program` the list of values and
        bind them to the `MOD`. See also https://docs.chia.net/guides/chialisp-currying

        Returns a program with the given values bound.
        """
        return self.to(self.curry_treehasher.curry(self, *args))

    def uncurry(self) -> Tuple[Program, Optional[List[Program]]]:
        """
        uncurry the given program

        returns `mod, [arg1, arg2, ...]`

        if the program is not a valid curry, return `self, NULL`

        This distinguishes it from the case of a valid curry of 0 arguments
        (which is rather pointless but possible), which returns `self, []`
        """
        mod, args = self.curry_treehasher.uncurry(self)
        p_args = args if args is None else [self.to(_) for _ in args]
        return self.to(mod), p_args

    def curry_hash(self, *args: bytes) -> bytes:
        """
        Return a puzzle hash that would be created if you curried this puzzle
        with arguments that have the given hashes.

        In other words,

        ```
        c1 = self.curry(arg1, arg2, arg3).tree_hash()
        c2 = self.curry_hash(arg1.tree_hash(), arg2.tree_hash(), arg3.tree_hash())
        assert c1 == c2  # they will be the same
        ```

        This looks useless to the unitiated, but sometimes you'll need a puzzle
        hash where you don't actually know the contents of a clvm subtree -- just its
        hash. This lets you calculate the puzzle hash with hidden information.
        """
        curry_treehasher = self.curry_treehasher
        quoted_mod_hash = curry_treehasher.calculate_hash_of_quoted_mod_hash(
            self.tree_hash()
        )
        return curry_treehasher.curry_and_treehash(quoted_mod_hash, *args)

    def as_int(self) -> Optional[int]:
        v = self.atom
        if v is None:
            return v
        return int_from_bytes(v)

    def as_iter(self) -> Iterator[Program]:
        v = self
        while v.pair:
            yield v.pair[0]
            v = v.pair[1]


NULL_PROGRAM = Program.fromhex("80")
ONE_PROGRAM = Program.fromhex("01")



from typing import Callable, List, Optional, Tuple

from .tree_hash import shatree_atom, shatree_pair

deserialize_as_tree: Optional[
    Callable[[bytes, bool], Tuple[List[Tuple[int, int, int]], Optional[List[bytes]]]]
]

try:
    from clvm_rs.clvm_rs import deserialize_as_tree
except ImportError:
    deserialize_as_tree = None


MAX_SINGLE_BYTE = 0x7F
CONS_BOX_MARKER = 0xFF


# ATOM: serialize_offset, serialize_end, atom_offset
# PAIR: serialize_offset, serialize_end, right_index

Triple = Tuple[int, int, int]
DeserOp = Callable[[bytes, int, List[Triple], List], int]


def deserialize_as_tuples(
    blob: bytes, cursor: int, calculate_tree_hash: bool
) -> Tuple[List[Tuple[int, int, int]], Optional[List[bytes]]]:
    if deserialize_as_tree:
        try:
            tree, hashes = deserialize_as_tree(blob, calculate_tree_hash)
        except OSError as ex:
            raise ValueError(ex)
        return tree, hashes

    def save_cursor(
        index: int,
        blob: bytes,
        cursor: int,
        obj_list: List[Triple],
        op_stack: List[DeserOp],
    ) -> int:
        blob_index = obj_list[index][0]
        assert blob[blob_index] == 0xFF
        v0 = obj_list[index][0]
        v2 = obj_list[index][2]
        obj_list[index] = (v0, cursor, v2)
        if calculate_tree_hash:
            left_hash = tree_hash_list[index + 1]
            hash_index = obj_list[index][2]
            right_hash = tree_hash_list[hash_index]
            tree_hash_list[index] = shatree_pair(left_hash, right_hash)
        return cursor

    def save_index(
        index: int,
        blob: bytes,
        cursor: int,
        obj_list: List[Triple],
        op_stack: List[DeserOp],
    ) -> int:
        e = obj_list[index]
        obj_list[index] = (e[0], e[1], len(obj_list))
        return cursor

    def parse_obj(
        blob: bytes, cursor: int, obj_list: List[Triple], op_stack: List[DeserOp]
    ) -> int:
        if cursor >= len(blob):
            raise ValueError("bad encoding")

        if blob[cursor] == CONS_BOX_MARKER:
            index = len(obj_list)
            obj_list.append((cursor, 0, 0))
            op_stack.append(lambda *args: save_cursor(index, *args))
            op_stack.append(parse_obj)
            op_stack.append(lambda *args: save_index(index, *args))
            op_stack.append(parse_obj)
            if calculate_tree_hash:
                tree_hash_list.append(b"")
            return cursor + 1
        atom_offset, new_cursor = _atom_size_from_cursor(blob, cursor)
        my_hash = None
        if calculate_tree_hash:
            my_hash = shatree_atom(blob[cursor + atom_offset:new_cursor])
            tree_hash_list.append(my_hash)
        obj_list.append((cursor, new_cursor, atom_offset))
        return new_cursor

    obj_list: List[Triple] = []
    tree_hash_list: List[bytes] = []
    op_stack: List[DeserOp] = [parse_obj]
    while op_stack:
        f = op_stack.pop()
        cursor = f(blob, cursor, obj_list, op_stack)
    return obj_list, tree_hash_list if calculate_tree_hash else None


def _atom_size_from_cursor(blob, cursor) -> Tuple[int, int]:
    # return `(size_of_prefix, cursor)`
    b = blob[cursor]
    if b == 0x80:
        return 1, cursor + 1
    if b <= MAX_SINGLE_BYTE:
        return 0, cursor + 1
    bit_count = 0
    bit_mask = 0x80
    while b & bit_mask:
        bit_count += 1
        b &= 0xFF ^ bit_mask
        bit_mask >>= 1
    size_blob = bytes([b])
    if bit_count > 1:
        size_blob += blob[cursor + 1:cursor + bit_count]
    size = int.from_bytes(size_blob, "big")
    new_cursor = cursor + size + bit_count
    if new_cursor > len(blob):
        raise ValueError("end of stream")
    return bit_count, new_cursor



from __future__ import annotations
from typing import Dict

from .casts import CastableType
from .clvm_storage import CLVMStorage


def replace(program: CLVMStorage, **kwargs: CastableType) -> CastableType:
    # if `kwargs == {}` then `return program` unchanged
    if len(kwargs) == 0:
        return program

    if "" in kwargs:
        if len(kwargs) > 1:
            raise ValueError("conflicting paths")
        return kwargs[""]

    # we've confirmed that no `kwargs` is the empty string.
    # Now split `kwargs` into two groups: those
    # that start with `f` and those that start with `r`

    args_by_prefix: Dict[str, Dict[str, CastableType]] = dict(f={}, r={})
    for k, v in kwargs.items():
        c = k[0]
        if c not in "fr":
            msg = f"bad path containing {c}: must only contain `f` and `r`"
            raise ValueError(msg)
        args_by_prefix[c][k[1:]] = v

    pair = program.pair
    if pair is None:
        raise ValueError("path into atom")

    # recurse down the tree
    new_f = replace(pair[0], **args_by_prefix.get("f", {}))
    new_r = replace(pair[1], **args_by_prefix.get("r", {}))

    return (new_f, new_r)



from .clvm_storage import CLVMStorage
from .de import deserialize_as_tuples

from typing import List, Optional, Tuple, Union


class CLVMTree(CLVMStorage):
    """
    This object conforms with the `CLVMStorage` protocol. It's optimized for
    deserialization, and keeps a reference to the serialized blob and to a
    list of tuples of integers, each of which corresponds to a subtree.

    It turns out every atom serialized to a blob contains a substring that
    exactly matches that atom, so it ends up being not very wasteful to
    simply use the blob for atom storage (especially if it's a `memoryview`,
    from which you can take substrings without copying). Additionally, the
    serialization for every object in the tree exactly corresponds to a
    substring in the blob, so by clever caching we can very quickly generate
    serializations for any subtree.

    The deserializer iterates through the blob and caches a triple of
    integers for each subtree: the first two integers represent the
    `(start_offset, end_offset)` within the blob that corresponds to the
    serialization of that object. You can check the contents of
    `blob[start_offset]` to determine if the object is a pair (in which case
    that byte is 0xff) or an atom (anything else). For a pair, the third
    number corresponds to the index of the array that is the "rest" of the
    pair (the "first" is always this object's index plus one, so we don't
    need to save that); for an atom, the third number corresponds to an
    offset of where the atom's binary data is relative to
    `blob[start_offset]` (so the atom data is at `blob[triple[0] +
    triple[2]:triple[1]]`)

    Since each `CLVMTree` subtree keeps a reference to the original
    serialized data and the list of tuples, no memory is released until all
    objects in the tree are garbage-collected. This happens pretty naturally
    in well-behaved python code.
    """

    # cached value of lazily-created child `CLVMStorage` objects
    # created on call to `.pair`
    _pair: Optional[Tuple["CLVMStorage", "CLVMStorage"]]

    # the serialized form, which contains the template for all the atoms
    blob: Union[memoryview, bytes]

    # the memoized boundary list for all clvm objects in the tree
    # this list is shared by *all* instances of sub-objects of the
    # same tree
    # [int, int, int] => start_offset, end_offset, atom_start_or_pair_rest
    #   deserialization of this object is blob[start_offset, end_offset]
    # We look at blob[start_offset] and if 0xff, this is a pair, otherwise atom
    # if atom, atom is blob[atom_start_or_pair_rest:end_offset]
    # if pair, children have `index` start_offset+1 and atom_start_or_pair_rest
    int_tuples: List[Tuple[int, int, int]]

    # which of the clvm objects in the tree this instance represents
    index: int  # index into `int_tuples`

    # the sha256 tree hashes that are optionally created on deserialization
    tree_hashes: Optional[List[bytes]]

    @classmethod
    def from_bytes(cls, blob: bytes, calculate_tree_hash: bool = True) -> "CLVMTree":
        int_tuples, tree_hashes = deserialize_as_tuples(
            blob, 0, calculate_tree_hash=calculate_tree_hash
        )
        return cls(memoryview(blob), int_tuples, tree_hashes, 0)

    def __init__(
        self,
        blob: Union[memoryview, bytes],
        int_tuples: List[Tuple[int, int, int]],
        tree_hashes: Optional[List[bytes]],
        index: int,
    ):
        self.blob = blob
        self.int_tuples = int_tuples
        self.tree_hashes = tree_hashes
        self.index = index
        if self.tree_hashes:
            self._cached_sha256_treehash = self.tree_hashes[index]
        start, end, atom_offset = self.int_tuples[self.index]
        if self.blob[start] == 0xFF:
            self.atom = None
        else:
            self.atom = bytes(self.blob[start + atom_offset:end])
            self._pair = None

    @property
    def pair(self) -> Optional[Tuple["CLVMStorage", "CLVMStorage"]]:
        if not hasattr(self, "_pair"):
            tuples, tree_hashes = self.int_tuples, self.tree_hashes
            start, end, right_index = tuples[self.index]
            # if `self.blob[start]` is 0xff, it's a pair
            assert self.blob[start] == 0xFF
            left = self.__class__(self.blob, tuples, tree_hashes, self.index + 1)
            right = self.__class__(self.blob, tuples, tree_hashes, right_index)
            self._pair = (left, right)
        return self._pair

    @property
    def _cached_serialization(self) -> bytes:
        start, end, _ = self.int_tuples[self.index]
        return self.blob[start:end]

    def __bytes__(self) -> bytes:
        return bytes(self._cached_serialization)

    def __str__(self) -> str:
        return bytes(self).hex()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"



import io
import unittest

from clvm_rs.program import Program
from clvm_rs.ser import atom_to_byte_iterator


TEXT = b"the quick brown fox jumps over the lazy dogs"


class InfiniteStream(io.TextIOBase):
    def __init__(self, b):
        self.buf = b

    def read(self, n):
        ret = b""
        while n > 0 and len(self.buf) > 0:
            ret += self.buf[0:1]
            self.buf = self.buf[1:]
            n -= 1
        ret += b" " * n
        return ret


class LargeAtom:
    def __len__(self):
        return 0x400000001


class SerializeTest(unittest.TestCase):
    def check_serde(self, s):
        v = Program.to(s)
        b = bytes(v)
        f = io.BytesIO()
        v.stream(f)
        b1 = f.getvalue()
        self.assertEqual(b, b1)
        v1 = Program.parse(io.BytesIO(b))
        if v != v1:
            print("%s: %d %s %s" % (v, len(b), b, v1))
            breakpoint()
            b = bytes(v)
            v1 = Program.parse(io.BytesIO(b))
        self.assertEqual(v, v1)

    def test_zero(self):
        v = Program.to(b"\x00")
        self.assertEqual(bytes(v), b"\x00")

    def test_empty(self):
        v = Program.to(b"")
        self.assertEqual(bytes(v), b"\x80")

    def test_empty_string(self):
        self.check_serde(b"")

    def test_single_bytes(self):
        for _ in range(256):
            self.check_serde(bytes([_]))

    def test_short_lists(self):
        self.check_serde([])
        for _ in range(0, 2048, 8):
            for size in range(1, 5):
                self.check_serde([_] * size)

    def test_cons_box(self):
        self.check_serde((0, 0))
        self.check_serde((0, [1, 2, 30, 40, 600, ([], 18)]))
        self.check_serde((100, (TEXT, (30, (50, (90, (TEXT, TEXT + TEXT)))))))

    def test_long_blobs(self):
        text = TEXT * 300
        for _, t in enumerate(text):
            t1 = text[:_]
            self.check_serde(t1)

    def test_blob_limit(self):
        with self.assertRaises(ValueError):
            next(atom_to_byte_iterator(LargeAtom()))

    def test_very_long_blobs(self):
        for size in [0x40, 0x2000, 0x100000, 0x8000000]:
            count = size // len(TEXT)
            text = TEXT * count
            assert len(text) < size
            self.check_serde(text)
            text = TEXT * (count + 1)
            assert len(text) > size
            self.check_serde(text)

    def test_very_deep_tree(self):
        blob = b"a"
        for depth in [10, 100, 1000, 10000, 100000]:
            s = Program.to(blob)
            for _ in range(depth):
                s = Program.to((s, blob))
            self.check_serde(s)

    def test_deserialize_empty(self):
        bytes_in = b""
        with self.assertRaises(ValueError):
            Program.from_bytes(bytes_in)
        with self.assertRaises(ValueError):
            Program.parse(io.BytesIO(bytes_in))

    def test_deserialize_truncated_size(self):
        # fe means the total number of bytes in the length-prefix is 7
        # one for each bit set. 5 bytes is too few
        bytes_in = b"\xfe    "
        with self.assertRaises(ValueError):
            Program.from_bytes(bytes_in)
        with self.assertRaises(ValueError):
            Program.parse(io.BytesIO(bytes_in))

    def test_deserialize_truncated_blob(self):
        # this is a complete length prefix. The blob is supposed to be 63 bytes
        # the blob itself is truncated though, it's less than 63 bytes
        bytes_in = b"\xbf   "

        with self.assertRaises(ValueError):
            Program.from_bytes(bytes_in)
        with self.assertRaises(ValueError):
            Program.parse(io.BytesIO(bytes_in))

    def test_deserialize_large_blob(self):
        # this length prefix is 7 bytes long, the last 6 bytes specifies the
        # length of the blob, which is 0xffffffffffff, or (2^48 - 1)
        # we don't support blobs this large, and we should fail immediately
        # when exceeding the max blob size, rather than trying to read this
        # many bytes from the stream
        bytes_in = b"\xfe" + b"\xff" * 6

        with self.assertRaises(ValueError):
            Program.parse(InfiniteStream(bytes_in))

    def test_repr_clvm_tree(self):
        with self.assertRaises(ValueError):
            Program.fromhex("ff8085")

        o = Program.fromhex("ff808185")
        self.assertEqual(repr(o._unwrapped_pair[0]), "<CLVMTree: 80>")
        self.assertEqual(repr(o._unwrapped_pair[1]), "<CLVMTree: 8185>")

    def test_bad_blob(self):
        self.assertRaises(ValueError, lambda: Program.fromhex("ff"))
        f = io.BytesIO(bytes.fromhex("ff"))
        self.assertRaises(ValueError, lambda: Program.parse(f))

    def test_large_atom(self):
        s = "foo" * 100
        p = Program.to(s)
        blob = bytes(p)
        p1 = Program.from_bytes(blob)
        self.assertEqual(p, p1)

    def test_too_large_atom(self):
        self.assertRaises(ValueError, lambda: Program.fromhex("fc"))
        self.assertRaises(ValueError, lambda: Program.fromhex("fc8000000000"))



"""
This is an implementation of `sha256_treehash`, used to calculate
puzzle hashes in clvm.

This implementation goes to great pains to be non-recursive so we don't
have to worry about blowing out the python stack.
"""

from hashlib import sha256
from typing import Callable, List, Tuple, cast

from .clvm_storage import CLVMStorage


OP_STACK_F = Callable[[List[CLVMStorage], List[bytes], List["OP_STACK_F"]], None]


class Treehasher:
    """
    `Treehasher` performs the standard sha256tree hashing in a non-recursive
    way so that extremely large objects don't blow out the python stack.

    We also force a `_cached_sha256_treehash` into the hashed sub-objects
    whenever possible so that taking the hash of the same sub-tree is
    more efficient in future.
    """

    atom_prefix: bytes
    pair_prefix: bytes
    cache_hits: int

    def __init__(self, atom_prefix: bytes, pair_prefix: bytes):
        self.atom_prefix = atom_prefix
        self.pair_prefix = pair_prefix
        self.cache_hits = 0

    def shatree_atom(self, atom: bytes) -> bytes:
        s = sha256()
        s.update(self.atom_prefix)
        s.update(atom)
        return s.digest()

    def shatree_pair(self, left_hash: bytes, right_hash: bytes) -> bytes:
        s = sha256()
        s.update(self.pair_prefix)
        s.update(left_hash)
        s.update(right_hash)
        return s.digest()

    def sha256_treehash(self, clvm_storage: CLVMStorage) -> bytes:
        def handle_obj(
            obj_stack: List[CLVMStorage],
            hash_stack: List[bytes],
            op_stack: List[OP_STACK_F],
        ) -> None:
            obj = obj_stack.pop()
            r = getattr(obj, "_cached_sha256_treehash", None)
            if r is not None:
                self.cache_hits += 1
                hash_stack.append(r)
                return
            elif obj.atom is not None:
                r = shatree_atom(obj.atom)
                hash_stack.append(r)
                try:
                    setattr(obj, "_cached_sha256_treehash", r)
                except AttributeError:
                    pass
            else:
                pair = cast(Tuple[CLVMStorage, CLVMStorage], obj.pair)
                p0, p1 = pair
                obj_stack.append(obj)
                obj_stack.append(p0)
                obj_stack.append(p1)
                op_stack.append(handle_pair)
                op_stack.append(handle_obj)
                op_stack.append(handle_obj)

        def handle_pair(
            obj_stack: List[CLVMStorage],
            hash_stack: List[bytes],
            op_stack: List[OP_STACK_F],
        ) -> None:
            p0 = hash_stack.pop()
            p1 = hash_stack.pop()
            r = shatree_pair(p0, p1)
            hash_stack.append(r)
            obj = obj_stack.pop()
            try:
                setattr(obj, "_cached_sha256_treehash", r)
            except AttributeError:
                pass

        obj_stack: List[CLVMStorage] = [clvm_storage]
        op_stack: List[OP_STACK_F] = [handle_obj]
        hash_stack: List[bytes] = []
        while len(op_stack) > 0:
            op: OP_STACK_F = op_stack.pop()
            op(obj_stack, hash_stack, op_stack)
        return hash_stack[0]


CHIA_TREE_HASH_ATOM_PREFIX = bytes.fromhex("01")
CHIA_TREE_HASH_PAIR_PREFIX = bytes.fromhex("02")
CHIA_TREEHASHER = Treehasher(CHIA_TREE_HASH_ATOM_PREFIX, CHIA_TREE_HASH_PAIR_PREFIX)

sha256_treehash = CHIA_TREEHASHER.sha256_treehash
shatree_atom = CHIA_TREEHASHER.shatree_atom
shatree_pair = CHIA_TREEHASHER.shatree_pair



import pytest

from clvm_rs import Program
from clvm_rs.chia_dialect import CHIA_DIALECT
from clvm_rs.curry_and_treehash import CurryTreehasher

CHIA_CURRY_TREEHASHER = CurryTreehasher(CHIA_DIALECT)
curry_and_treehash = CHIA_CURRY_TREEHASHER.curry_and_treehash
calculate_hash_of_quoted_mod_hash = (
    CHIA_CURRY_TREEHASHER.calculate_hash_of_quoted_mod_hash
)


def test_curry_and_treehash() -> None:
    arbitrary_mod = Program.fromhex("ff10ff02ff0580")  # `(+ 2 5)`
    arbitrary_mod_hash = arbitrary_mod.tree_hash()

    # we don't really care what `arbitrary_mod` is. We just need some code

    quoted_mod_hash = calculate_hash_of_quoted_mod_hash(arbitrary_mod_hash)
    exp_hash = "9f487f9078d4b215e0cbe2cbdd21215ad6ed8e894ae00d616751e0efdccb25a9"
    assert quoted_mod_hash == bytes.fromhex(exp_hash)

    for v in range(500):
        args = [v, v * v, v * v * v]
        # we don't really care about the arguments either
        puzzle = arbitrary_mod.curry(*args)
        puzzle_hash_via_curry = puzzle.tree_hash()
        hashed_args = [Program.to(_).tree_hash() for _ in args]
        puzzle_hash_via_f = curry_and_treehash(quoted_mod_hash, *hashed_args)
        assert puzzle_hash_via_curry == puzzle_hash_via_f
        puzzle_hash_via_m = arbitrary_mod.curry_hash(*hashed_args)
        assert puzzle_hash_via_curry == puzzle_hash_via_m


def test_bad_parameter() -> None:
    arbitrary_mod = Program.fromhex("ff10ff02ff0580")  # `(+ 2 5)`
    with pytest.raises(ValueError):
        arbitrary_mod.curry_hash(b"foo")
